"""
Copyright (c) 2024 Carlos Alberto da Costa Filho. All rights reserved.

Curvelets: Python package for curvelet transforms.
"""

from __future__ import annotations

from ._version import version as __version__

__all__ = ["__version__", "numpy", "reference"]

from . import numpy, reference
